function redis_connect()
    local redis = require "nginx.redis"
    local red = redis:new()

    red:set_timeout(1000)

    local auth_redis_host = ngx.var.auth_redis_host or '127.0.0.1'
    local auth_redis_port = ngx.var.auth_redis_port or 6379

    local ok, err = red:connect(auth_redis_host, auth_redis_port)
    if not ok then
        ngx.log(ngx.ERR, "failed to connect to redis server: ", err)
        ngx.exit(ngx.HTTP_INTERNAL_SERVER_ERROR)
        return
    end

    return red
end

function is_authenticate()
    local uid = ngx.var.uid_got
    if uid == nil then
        return false
    end

    if uid:find('=') == nil then
        ngx.log(ngx.ERR, "unable to find uid: " .. uid)
        ngx.exit(ngx.HTTP_INTERNAL_SERVER_ERROR)
    end

    local red = redis_connect()

    local res, err = red:exists(uid)
    if not res then
        if err then
            ngx.log(ngx.ERR, "redis error: " .. err)
        end
        ngx.exit(ngx.HTTP_INTERNAL_SERVER_ERROR)
        return
    end

    return res == 1
end

function authenticate()
    local cjson    = require "cjson"
    local header   = ngx.req.get_headers()['Authorization']

    if header == nil or header:find(' ') == nil then
        return false
    end

    local divider = header:find(' ')
    if header:sub(0, divider - 1) ~= 'Basic' then
        return false
    end

    local auth = ngx.decode_base64(header:sub(divider + 1))
    if auth == nil or auth:find(':') == nil then
        return false
    end

    divider        = auth:find(':')
    local username = auth:sub(0, divider - 1)
    local payload  = cjson.encode({password = auth:sub(divider + 1)})

    local res      = ngx.location.capture("/thallium/" .. string.gsub(username, "@", "-at-"),
                                          { method = ngx.HTTP_POST,
                                            body   = payload })

    if res.status ~= ngx.HTTP_OK or res.body == nil then
        return false
    end

    local body     = cjson.decode(res.body)

    if body['auth'] == nil or body['auth']['policies'] == nil then
        return false
    end

    local policies   = cjson.decode(ngx.var.THALLIUM_POLICIES)
    local has_policy = false

    if (type(policies) ~= 'table') then
        ngx.log(ngx.ERR, "invalid thallium policies")
        ngx.exit(ngx.HTTP_INTERNAL_SERVER_ERROR)
        return false
    end

    for index, value in ipairs(body['auth']['policies']) do
        if policies[value] ~= nil then
            has_policy = true
            break
        end
    end

    if has_policy == false then
        return false
    end

    local uid = ngx.var.uid_set or ngx.var.uid_got
    if uid == nil then
        return false
    end

    local uinfo         = {}
    uinfo['uid']        = uid
    uinfo['username']   = username
    uinfo['started_at'] = os.time()
    uinfo['policies']   = body['auth']['policies']

    local red = redis_connect()
    local res, err = red:setex(uid, ngx.var.auth_expire or 1440, cjson.encode(uinfo))
    if not res then
        if err then
            ngx.log(ngx.ERR, "redis error: " .. err)
        end
        ngx.exit(ngx.HTTP_INTERNAL_SERVER_ERROR)
        return
    end

    return true
end

if not is_authenticate() and not authenticate() then
    ngx.header.content_type = 'text/plain'
    ngx.header.www_authenticate = 'Basic realm="restricted area"'
    ngx.status = ngx.HTTP_UNAUTHORIZED
    ngx.say("401 Unauthorized")
end
